/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1992, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

 * This version of the delay routine uses DOS calls instead of BIOS
 * calls to make it compatable with Japanese machines

  Modified:	By:		Reason:
  ---------	---		-------

  25-jul-92	F.W.Crigger	Rewritten to use straight DOS int 21h calls
				which should make it more portable and work
				in OS/2 DOS Boxes, etc.
  20-aug-92	J.B.Schueler	Initial implementation in C based on assembly
				code from delay.asm (386 version)
  15-feb-93	A.Gupta		Netware support
--------------------------------------------------------Version-11.0---------
  30-may-95	Greg Bentz	variety.h and dll linkage
  18-sep-95	T. Schiller	Rewritten to reduce initialization delay.
  19-jun-96	M. Hildebrand	Renamed __nonIBM to __is_nonIBM to avoid
				namespace conflict with __NonIBM global.
  12-jun-97	Alex Brodsky	Some old NEC98 models have a broken
  				hundred of a second clock which caused
				program hangs.
 */
#include "variety.h"
#include "rtinit.h"

/*
 * Using interupt 21 service 2C, we get the time from DOS
 * in CX and DX as follows:
 *
 *	CH = hours
 *	CL = minutes
 *	DH = seconds
 *	DL = seconds / 100
 */
#ifdef __386__
    extern unsigned GetDosHSecs(void);
    #pragma aux GetDosHSecs =	\
	"mov    ah, 2ch"		\
	"xor    edx, edx"		\
	"int    21h"		\
	"xor    dh, dh"		\
	value [edx]			\
	modify [eax ecx edx];
    
    extern unsigned GetDosSecs(void);
    #pragma aux GetDosSecs =	\
	"mov    ah, 2ch"		\
	"xor    edx, edx"		\
	"int    21h"		\
	"xor    dl, dl"		\
	value [edx]			\
	modify [eax ecx edx];
#else
    extern unsigned GetDosHSecs(void);
    #pragma aux GetDosHSecs =	\
	"mov    ah, 2ch"		\
	"int    21h"		\
	"xor    dh, dh"		\
	value [dx]			\
	modify [ax cx dx];
    
    extern unsigned GetDosSecs(void);
    #pragma aux GetDosSecs =	\
	"mov    ah, 2ch"		\
	"int    21h"		\
	"xor    dl, dl"		\
	value [dx]			\
	modify [ax cx dx];
#endif
    
extern int __is_nonIBM( void );
    
static long loop_count = 0L;
static int  broken_nec = 0;

void __delay_init(void)
{
    unsigned last;
    unsigned current;
    long     total = 10L;
    long     left;
    long     hsec;
    long     sec;
    long     nsec;

    /*
     * On NEC machines, DL contains the number of milliseconds with the
     * oddity that it only counts from 0 to 99 and then sits on 99 for the
     * remaining 9/10th of a second.
     */
    if (__is_nonIBM())
    {
	total = 50L; // 1/20th of a second

	/*
	 * Since we only have accuracy in the first 1/10 of the second, if
	 * we are getting close to hitting 99 milliseconds, then we need to
	 * wait until the next second.
	 */
	hsec = GetDosHSecs();
	if (hsec >= 25)
	{
	    last = GetDosSecs();
	    while (GetDosSecs() == last)
		;
	} else if( hsec == 0 ) {
	    // On some Old NEC98 hsec will always be 0 and this causes the
	    // initialization to hang
	    sec = GetDosSecs();
	    do {
		hsec = GetDosHSecs();
	        nsec = GetDosSecs();
	    } while( !hsec && ( sec == nsec ) );

	    if( !hsec ) {
		// Well, hsec is broken so we need to calculate roughly 1
		// millisecond to provide equivalent support (or try)
		do {
		    sec = GetDosSecs();
		    loop_count++;
		} while( sec == nsec );
		broken_nec = 1;
		loop_count /= 1000;
		return;
	    }
	}
    }

    last = GetDosHSecs();
    while ((current = GetDosHSecs()) == last)
	;

    for (loop_count = 0L, left = total; left > 0L; ++loop_count)
    {
	last    = current;
	current = GetDosHSecs();

	if (current == last)
	    continue;
	else if (current > last)
	    left -= (current - last);
	else
	    left -= (100L - last + current);
    } /* for */

    /*
     * Since we used only 1/20th of a second accuracy on NEC machines,
     * multiply the result by two to make the delay routine a little easier.
     */
    if (__is_nonIBM())
	loop_count <<= 1;
} /* __delay_init() */


AXI(__delay_init, INIT_PRIORITY_LIBRARY)


void delay(unsigned milliseconds)
{
    unsigned last;
    unsigned current = GetDosHSecs();
    long     count   = (milliseconds * loop_count + 99) / 100;

    if (milliseconds == 0)
	return;

    if( broken_nec ) {
	// note: this will break for large millisecond counts.
	// but its better than nothing
	for( count = milliseconds * loop_count; count > 0; count-- ) {
	    // To maintain same counting time
	    last = GetDosSecs();
	}
	return;
    }
	
    for (; count > 0; --count)
    {
	/*
	 * Everything beyond this point exists only to give us the
	 * same number of operations that we used in the original
	 * calibration.
	 */
	last    = current;
	current = GetDosHSecs();

	if (current == last)
	    continue;
	else if (current > last)
	    current -= (current - last);
	else if (current < last)
	    current -= (100L - last + current);
    } /* for */
} /* delay() */
