/*
**********************************************************************
*       Copyright by WATCOM International Corporation, 1987, 1992.   *
*       All rights reserved. No part of this software may be         *
*       reproduced in any form or by any means - graphic, electronic,*
*       mechanical or otherwise, including, without limitation,      *
*       photocopying, recording, taping or information storage and   *
*       retrieval systems - except with the written permission of    *
*       WATCOM International Corporation.                            *
**********************************************************************

    Complex class functions and operators

   The formula for some of these algorithms was found in:
     "Complex Variables and Applications, 2nd Edition",
     Ruel V. Churchill,
     McGraw-Hill, 1960
   (Look in Jim Welch's office)
   These formulae include:
     acos, asin, atan,
     cosh, sinh, tanh

    Modified    By              Reason
    ========    ==              ======
    92/01/09    Steve McDowell  Initial implementation.
    92/10/09	Greg Bentz	Cleanup.
    93/02/16	Greg Bentz	Removed some numerical instability.
    93/10/19    Raymond Tang    Split into separate files.
    95/05/19	Greg Bentz	dll linkage
*/

#include "variety.h"
#include <complex>
#include "cplx.h"

_WPRTLINK Complex atan( const Complex &z ) {
/*****************************************/
// Arctan of a complex number.
// From "Complex Variables and Applications" (pg. 63).
//    atan( z ) = i/2 * log( (i + z)/(i - z) )
//              = i/2 * log( (i + (z.r + z.i*i))/(i - (z.r + z.i*i)) )
//              = i/2 * log( (z.r + (1+z.i)*i)/(-z.r + (1-z.i)*i) )
//
// let a = z.r, b = z.i
//			     (a + (1 + b)*i)/(-a + (1 - b)*i)
//
// or:   a+(1+b)*i     -a-(1-b)*i     - a^2 + 1 - b^2 - (1 - b + 1 + b )*a*i
//     ------------ * ------------ = ----------------------------------------
//      -a+(1-b)*i     -a-(1-b)*i              a^2 + b^2 - 2*b + 1
//
//                                    1 - a^2 - b^2 - 2*a*i
//                                 = -----------------------
//                                     a^2 + b^2 - 2*b + 1
//
// Note: a^2 + b^2 == (a+b)^2 - 2ab
//                                    1 - (a+b)^2 + 2*a*b - 2*a*i
//                                 = -----------------------------
//                                     (a+b)^2 - 2*a*b - 2*b + 1
//
    dcomplex    value;
    double	apb2 = (z.real()+z.imag())*(z.real()+z.imag());
    double	_2ab = 2.0*z.real()*z.imag();

    value = _IF_C16Div( 1.0 - apb2 + _2ab, -2.0*z.real(),
                    apb2 - _2ab - 2.0*z.imag() + 1.0, 0.0 );
    value = _IF_CDLOG( value.realpart, value.imagpart );
    return Complex( -value.imagpart / 2, value.realpart / 2 );
}

