/*
**********************************************************************
*       Copyright by WATCOM International Corporation, 1987, 1994.   *
*       All rights reserved. No part of this software may be         *
*       reproduced in any form or by any means - graphic, electronic,*
*       mechanical or otherwise, including, without limitation,      *
*       photocopying, recording, taping or information storage and   *
*       retrieval systems - except with the written permission of    *
*       WATCOM International Corporation.                            *
**********************************************************************

EXC_PR -- entries to control exception searching

These entries are allocated on the stack by the run-time system.  They are
linked, in definition order, from the thread data area (exc_pr).

Their purpose is control searching and error detection while searching
the exceptions R/W blocks.  Essentially, they represent an additional
thread through some of the active R/W blocks, recording significant actions
that are in progress:
    - function exception specification failure
        - unexpected called
        - bad_exception thrown
    - construction of exception in allocation area
    - destruction during stack unwind
    - terminate called

The constructor links in the entry; the destructor will unlink it during
stack unwinding.

unexpected processing
---------------------
- only when user-defined routine to be called
- unexpected r/t routine links in an fnexc entry
- if we search to that point on a subsequent throw, then status is set to
  avoid examining R/W blocks until the point of the original fn-exc is reached

other processing
----------------
- designed to catch exceptions leaking out in bad circumstances
    - after unexpected (see above), terminate
    - dtor during stack unwind
    - ctor during copy to allocation area
- if we search to that point on a subsequent throw, special error handling
  will occur


Modified        By              Reason
--------        --              ------
92/10/02        Greg Bentz      Initial version, shipped as "alpha"
                Jim Randall
                Anthony Scian
                Jim Welch
93/05/01        Greg Bentz      Version 9.5 shipped
                Anthony Scian
                Jim Welch
93/08/31        Greg Bentz      Version 9.5 patch level A
                Anthony Scian
                Jim Welch
94/02/04        Greg Bentz      Version 9.5 patch level B
                Anthony Scian
                Jim Welch
94/03/21        Greg Bentz      Version 10.0 LA Version
                Anthony Scian
                Jim Welch
94/05/31        Greg Bentz      Version 10.0 GA Version
                Anthony Scian
                Jim Welch
94/08/31        Greg Bentz      Version 10.0 patch level A
                Anthony Scian
                Jim Welch

95/09/14        J.W.Welch       Defined
*/

#include "cpplib.h"
#include "rtexcept.h"
#include "exc_pr.h"

//***********************************************************************
// Implementation of _EXC_PR
// - basic entry
// - used as base for _EXC_PR_FNEXC, _EXC_PR_EXC
//***********************************************************************

_EXC_PR::_EXC_PR                // _EXC_PR CONSTRUCTOR
    ( _RTCTL* rtc               // - run-time control
    , RW_DTREG *rw              // - current read/write
    , EXCSTATE state )          // - state
    : _state( state )
    , _rw( rw )
    , _prev( rtc->thr->exc_pr )
    , _rtc( rtc )
#ifndef NDEBUG
    , _type( EXCPR_BASIC )
#endif
{
    rtc->thr->exc_pr = this;
    if( 0 == rw ) {
#ifdef RW_REGISTRATION
        _rw = FsTop();
#else
        _rw = CPPLIB( pd_top )();
#endif
    }
}


_EXC_PR::~_EXC_PR               // _EXC_PR DESTRUCTOR
    ( void )
{
    _rtc->thr->exc_pr = _prev;
}


//***********************************************************************
// Implementation of _EXC_PR_FREE
// - same semantics as _EXC_PR
// - also frees an exception saved in allocation area
//***********************************************************************


_EXC_PR_FREE::_EXC_PR_FREE      // _EXC_PR_FREE CONSTRUCTOR
    ( _RTCTL* rtc               // - run-time control
    , RW_DTREG *rw              // - current read/write
    , EXCSTATE state            // - state
    , ACTIVE_EXC* exc )         // - exception to be destructed
    : _EXC_PR( rtc, rw, state )
    , _exc( exc )
{
#ifndef NDEBUG
    _type = EXCPR_FREE;
#endif
}


_EXC_PR_FREE::~_EXC_PR_FREE     // _EXC_PR_FREE DESTRUCTOR
    ( void )
{
    if( NULL != _exc ) {
        ACTIVE_EXC* exc = _exc;
        _exc = 0;
        CPPLIB( free_exc )( _rtc, exc );
    }
}


//***********************************************************************
// Implementation of _EXC_PR_DTOR
// - same semantics as _EXC_PR
// - also destructs and frees an exception saved in allocation area
//***********************************************************************


_EXC_PR_DTOR::_EXC_PR_DTOR      // _EXC_PR_DTOR CONSTRUCTOR
    ( _RTCTL* rtc               // - run-time control
    , RW_DTREG *rw              // - current read/write
    , EXCSTATE state            // - state
    , ACTIVE_EXC* exc )         // - exception to be destructed
    : _EXC_PR_FREE( rtc, rw, state, exc )
{
#ifndef NDEBUG
    _type = EXCPR_DTOR;
#endif
}


_EXC_PR_DTOR::~_EXC_PR_DTOR     // _EXC_PR_DTOR DESTRUCTOR
    ( void )
{
    ACTIVE_EXC* exc = _exc;
    CPPLIB( dtor_free_exc )( _exc, _rtc );
    _exc = 0;
}


//***********************************************************************
// Implementation of _EXC_PR_FNC
// - entry signifying that unexpected() has been invoked
// - modified when unexpected() throws/rethrows something not in
//   original fn-exc specification
//***********************************************************************


_EXC_PR_FNEXC::_EXC_PR_FNEXC    // _EXC_PR_FNEXC CONSTRUCTOR
    ( _RTCTL* rtc               // - run-time control
    , RW_DTREG *rw              // - current read/write
    , RW_DTREG *fnexc_skip      // - skipping read/write
    , ACTIVE_EXC* exc )         // - exception to be destructed
    : _EXC_PR_DTOR( rtc, rw, EXCSTATE_UNEXPECTED, exc )
    , _fnexc_skip( fnexc_skip == 0 ? _rw : fnexc_skip )
{
#ifndef NDEBUG
    _type = EXCPR_FNEXC;
#endif
}


_EXC_PR_FNEXC::~_EXC_PR_FNEXC   // _EXC_PR_FNEXC DESTRUCTOR
    ( void )
{
}
