
/*
**********************************************************************
*       Copyright by WATCOM International Corporation, 1987, 1995.   *
*       All rights reserved. No part of this software may be         *
*       reproduced in any form or by any means - graphic, electronic,*
*       mechanical or otherwise, including, without limitation,      *
*       photocopying, recording, taping or information storage and   *
*       retrieval systems - except with the written permission of    *
*       WATCOM International Corporation.                            *
**********************************************************************

STRSRCH -- string searcher

This is an implementation of the Boyer-Moore Algorithm for
string searching, found in "Algorithms in C" by Robert Sedgewick,
Addison-Wesley, 1990, pages 287-288.
It also required explanation from Anthony Scian.
Since the algorithm requires a 256 entry table for the search and
needs to be able to store indices into the pattern in each entry,
a pattern longer than 255 would require the table to be an array of
"short" rather than "char". It was felt that this was excessive, so
a pattern longer than 255 is broken into two parts. The first part
is used to build the table, and any potential matches found using the
the BM algorithm are then completed using memcmp.

This algorithm could be improved as follows:
 For patterns > 255 characters, the SkipTable would have to be changed
 to use "unsigned int" rather than "unsigned char". Using alloca() to
 grab the table would be preferable to forcing the a stack size of
 256*sizeof( int ). Conditional code would have to be added for all
 references to SkipTable, in order to support patterns bigger and smaller
 than 255. If alloca() can't even get a 256 byte table, then a dumb
 search should be used. If the pattern > 255 characters but alloca()
 won't give more than 256, then the algorithm used here could be used.
 Therefore, there would be 4 cases:
     (1) pattern > 255 and alloca() gives a 256 * sizeof( int ) array
         for SkipTable. Use the current algorithm without the "extra_patlen"
         stuff.
     (2) pattern > 255 and alloca() gives a 256-byte array for SkipTable.
         Use the current algorithm including the "extra_patlen" stuff.
     (3) pattern <= 255 and alloca() gives a 256-byte array for SkipTable.
         Use the current algorithm.
     (4) alloca() won't give even a 256-byte array for SkipTable.
         Implement a dumb search. Use memchr to find the first character,
         then use memcmp to test the rest.

Modified        By              Reason
--------        --              ------
92/03/12        Steve McDowell  Initial implementation.

92/10/02        Greg Bentz      Initial version, shipped as "alpha"
                Jim Randall
                Anthony Scian
                Jim Welch
93/05/01        Greg Bentz      Version 9.5 shipped
                Anthony Scian
                Jim Welch
93/08/31        Greg Bentz      Version 9.5 patch level A
                Anthony Scian
                Jim Welch
94/02/04        Greg Bentz      Version 9.5 patch level B
                Anthony Scian
                Jim Welch
94/03/21        Greg Bentz      Version 10.0 LA Version
                Anthony Scian
                Jim Welch
94/05/31        Greg Bentz      Version 10.0 GA Version
                Anthony Scian
                Jim Welch
94/08/31        Greg Bentz      Version 10.0 patch level A
                Anthony Scian
                Jim Welch

95/10/27        J.W.Welch       Moved to CPP directory
*/
#include "variety.h"
#include <string.h>
#include "strsrch.def"

#define NOT_FOUND (-1)


extern "C" long StringSearch
    ( const char *  str
    , unsigned int  strlen
    , const char *  pattern
    , unsigned int  patlen )
/**************************/
{

// If the pattern is of length 0, return 0:

    if( patlen == 0 ) {
        return( 0 );
    }

// If the pattern is longer than the string, indicate NOT_FOUND:

    if( patlen > strlen ) {
        return( NOT_FOUND );
    }

// If the pattern is of length 1, use memchr:

    if( patlen == 1 ) {
        const char *  loc;

        loc = (const char *)memchr( str, *pattern, strlen );
        return( loc == NULL ? NOT_FOUND : loc - str );
    }

// Use the Boyer-Moore algorithm:

    {
        auto unsigned char  SkipTable[256];
             unsigned int   extra_patlen;   // pattern size > 255
             unsigned int   str_index;
             unsigned int   pat_index;
             unsigned int   skip;

// First, construct the SkipTable. For each character found in the pattern,
// the corresponding SkipTable entry contains the offset from the last
// character in the pattern of the LAST occurrence of that character
// in the pattern. For characters that do not occur in the pattern,
// the SkipTable entry contains the length of the pattern.

        if( patlen > 255 ) {
            extra_patlen = patlen - 255;
            patlen       = 255;
        } else {
            extra_patlen = 0;
        }
        memset( SkipTable, patlen, 256 );
        for( pat_index = 0; pat_index < patlen; ++pat_index ) {
            int c;

            c = (unsigned char) pattern[pat_index];
            SkipTable[c] = (unsigned char)(patlen - pat_index - 1);
        }

// Now, search for the string. Each comparison of "str" to "pattern" proceeds
// from right-to-left. When a mismatch occurs, the mismatched character is
// used to extract from SkipTable the amount to skip before attempting
// another match. If extra_patlen > 0, we use memcmp to compare the rest
// of the pattern.

        str_index = patlen;
        pat_index = patlen;
        for( ;; ) {
            if( pat_index == 0 ) {              // possible match?
                if( extra_patlen == 0 ) break;  // yes!
                if( memcmp( str + str_index + patlen,
                            pattern + patlen,
                            extra_patlen ) == 0 ) break;
                str_index += patlen + 1;       // advance only 1 char
                pat_index  = patlen;
            }
            --pat_index;
            --str_index;
            while( str[str_index] != pattern[pat_index] ) {
                skip       = SkipTable[str[str_index]];
                str_index += (patlen - pat_index > skip)
                                     ? (patlen - pat_index)
                                     : skip;
                if( str_index >= strlen - extra_patlen ) {
                    return( NOT_FOUND );
                }
                pat_index = patlen - 1;
            }
        }
        return( str_index );
    }
}
